import os
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from . import palettes
import errno

def colors_from_hue(data, hue, cmap):
    num_colors = len(data.groupby(hue))
    cm_subsection = np.linspace(0.2, 0.8, num_colors+2)
    cm_subsection = cm_subsection[1:num_colors+1]
    return [ cmap(v) for v in cm_subsection ]


# XXX: this method is a disaster
def plot(data, x, y, error=None, hue=None, markers=[None], linestyles=['-'], cmap=palettes.neon, label_lines=True):
    data = data.sort_values(x, ascending=True)

    if hue:
        if cmap is not None:
            colors = colors_from_hue(data, hue, cmap)

        for i,(key,grp) in enumerate(data.groupby(hue)):
            if label_lines:
                label = key
            else:
                label = ""
            marker = markers[i % len(markers)]
            linestyle = linestyles[i % len(linestyles)]
            if cmap is not None:
                color = colors[i]
            else:
                color = None
            if error is not None:
                plt.errorbar(list(grp[x]), list(grp[y]), yerr=grp[error], color=color, marker=marker, linestyle=linestyle, label=label)
            else:
                plt.plot(list(grp[x]), list(grp[y]), label=label, color=color, marker=marker, linestyle=linestyle)
    else:
        color = cmap(0.5)
        if error is not None:
            plt.errorbar(list(data[x]), list(data[y]), yerr=data[error], color=color, marker=markers[0], linestyle=linestyles[0])
        else:
            plt.plot(list(data[x]), list(data[y]), color=color, marker=markers[0], linestyle=linestyles[0])

    plt.ylabel(y)
    plt.xlabel(x)
    plt.gca().autoscale(tight=True)
    plt.gca().margins(y=0.1)

def stackplot(data, x, y, hue, cmap=palettes.neon):
    xs = np.array(data[x])
    yss = []
    labels = []
    for k,grp in data.groupby(hue):
        labels.append(k)
        grp_xs = grp[x].tolist()
        grp_ys = grp[y].tolist()
        ys = []
        for v in xs:
            if len(grp_xs) > 0 and grp_xs[0] == v:
                ys.append(grp_ys.pop(0))
                grp_xs.pop(0)
            else:
                if len(ys) == 0:
                    ys.append(0.)
                else:
                    ys.append(ys[-1])
        assert len(grp_xs) == 0
        assert len(grp_ys) == 0
        assert len(ys) == len(xs)
        yss.append(np.array(ys, dtype=float))

    if cmap is not None:
        colors = colors_from_hue(data, hue, cmap)
    else:
        colors = None

    plt.stackplot(xs, *yss, labels=labels, colors=colors)

    plt.ylabel(y)
    plt.xlabel(x)
    plt.gca().autoscale(tight=True)
    plt.gca().margins(y=0.1)

def scatter(data, x, y, hue=None, cmap=palettes.neon, markers=['o'], **kwargs):
    return plot(data=data,x=x,y=y, hue=hue,markers=markers, linestyles=[''],cmap=cmap, **kwargs)

def hist(data, x, hue=None, cmap=palettes.neon, **kwargs):
    """
    Plot a histogram from a dataframe column.

    If hue is provided, plot many overlayed histograms, one per value of the data[hue] column.

    Parameters
    ----------

      data: pandas.DataFrame
        dataframe to plot

      x: string
        column of data to plot

    Keyword Arguments
    -----------------

      hue: string
        column of dataframe to split on

      cmap
        colormap to use

      alpha: float
        opacity of histogram
    """
    if 'rwidth' not in kwargs:
        kwargs['rwidth'] = 0.92

    colors = colors_from_hue(data, hue, cmap)

    if hue:
        if 'alpha' not in kwargs:
            kwargs['alpha'] = 0.5

        for i,(k,grp) in enumerate(data.groupby(hue)):
            if cmap:
                color = colors[i]
            else:
                color = None
            print(color)

            plt.hist(grp[x], label=k, color=color, **kwargs)
    else:
        if 'color' not in kwargs:
            kwargs['color'] = cmap(0.5)

        plt.hist(data[x], **kwargs)



    plt.hist(x, **kwargs)

def cdf(data, *args, **kwargs):
    sorted_data = np.sort(data)
    sorted_data_cdf = np.arange(len(sorted_data))/float(len(sorted_data)) * 100 
    plt.plot(sorted_data, sorted_data_cdf, *args, **kwargs)
    
def savefig(filename, **kwargs):
    if not os.path.exists(os.path.dirname(filename)):
        try:
            os.makedirs(os.path.dirname(filename))
        except OSError as exc: # Guard against race condition
            if exc.errno != errno.EEXIST:
                raise

    plt.tight_layout(pad=0)
    plt.savefig(filename, **kwargs)



import matplotlib.pyplot as plt
from math import atan2,degrees
import numpy as np

def legend(*args, **kwargs):
    """
    plt.legend with a few extra location options:

    * inline: place legends on top of line
    * end: place legend at the right end of line

    Parameters
    ----------

      loc: string or int
        location to place the legend
    """
    if 'loc' in kwargs:
        if kwargs['loc'] == 'inline':
            del kwargs['loc']
            legend_inline(*args, **kwargs)
        elif kwargs['loc'] == 'end':
            del kwargs['loc']
            legend_end(*args, **kwargs)
        else:
            plt.legend(*args, **kwargs)
    else:
        plt.legend(*args, **kwargs)

def legend_end(*args, **kwargs):
    lines = plt.gca().get_lines()

    for line in lines:
        label = line.get_label()
        #Take only the lines which have labels other than the default ones
        if "_line" in label:
            continue
        x = line.get_xdata()[-1] + 0.05
        y = line.get_ydata()[-1]
        ax = line.axes
        ax.text(x,y,label,
                color=line.get_color(),
                bbox=dict(facecolor=ax.get_facecolor(),
                          edgecolor=ax.get_facecolor(),
                          x=x, y=y,
                          pad=2),
                **kwargs)

def legend_inline(xvals=None,xoffset=None,**kwargs):
    lines = plt.gca().get_lines()

    ax = lines[0].axes
    labLines = []
    labels = []

    #Take only the lines which have labels other than the default ones
    for line in lines:
        label = line.get_label()
        if "_line" not in label:
            labLines.append(line)
            labels.append(label)

    if xvals is None:
        xmin,xmax = ax.get_xlim()
        xvals = np.linspace(xmin,xmax,len(labLines)+2)[1:-1]

        if xoffset is not None:
            xvals += xoffset

    for line,x,label in zip(labLines,xvals,labels):
        label_line(line,x,label,**kwargs)

#Label line with line2D label data
def label_line(line,x,label=None,**kwargs):
    ax = line.axes
    xdata = line.get_xdata()
    ydata = line.get_ydata()

    #if (x < xdata[0]) or (x > xdata[-1]):
    #    print('x label location is outside data range!')
    #    return

    #Find corresponding y co-ordinate and angle of the
    ip = 1
    for i in range(len(xdata)):
        if x < xdata[i]:
            ip = i
            break

    y = ydata[ip-1] + (ydata[ip]-ydata[ip-1])*(x-xdata[ip-1])/(xdata[ip]-xdata[ip-1])

    if not label:
        label = line.get_label()

    trans_angle = 0

    #Set a bunch of keyword arguments
    if 'color' not in kwargs:
        kwargs['color'] = line.get_color()

    if ('horizontalalignment' not in kwargs) and ('ha' not in kwargs):
        kwargs['ha'] = 'center'

    if ('verticalalignment' not in kwargs) and ('va' not in kwargs):
        kwargs['va'] = 'center'

    if 'clip_on' not in kwargs:
        kwargs['clip_on'] = True

    if 'zorder' not in kwargs:
        kwargs['zorder'] = 2.5

    ax.text(x,y,label,rotation=trans_angle,
            bbox=dict(facecolor=ax.get_facecolor(),
                      edgecolor=ax.get_facecolor(),
                      pad=2),
            **kwargs)



"""
Styles for plots.

Mainly adds the cool, warn, and neon themes from `this post <https://blog.graphiq.com/finding-the-right-color-palettes-for-data-visualizations-fcd4e707a283>`_

.. testsetup:: *

   import plorts

"""

import matplotlib

def gradient_to_colormap(name, gradient):
    """
    Generates a matplotlib colormap from a gradient. 

    Plain white to black gradient

    >>> plorts.palettes.gradient_to_colormap('bw', [(0, "white"), (1, "black")])
    <matplotlib.colors.LinearSegmentedColormap object at ...>

    First 25% of gradient is white to blue, second 75% is blue to black

    >>> plorts.palettes.gradient_to_colormap('bw', [(0, "white"), (0.25, "#0000ff"), (1, "black")])
    <matplotlib.colors.LinearSegmentedColormap object at ...>
    
    Parameters
    ----------

      name
        name of the colormap to register with matplotlib

      gradient
        list of tuples specifying gradient

    Returns
    -------
    
      matplotlib.colors.LinearSegmentedColormap
    """
    gradient = [(pos,matplotlib.colors.to_rgb(name)) for pos,name in gradient]
    cmap = {"red":[], "green":[], "blue":[]}
    for pos,(r,g,b) in gradient:
        cmap["red"].append((pos,r,r))
        cmap["green"].append((pos,g,g))
        cmap["blue"].append((pos,b,b))
    return matplotlib.colors.LinearSegmentedColormap(name, cmap)


#: colorblind-friendly colors for use in axes.prop_cycle
colorblind = ["#0072B2", "#009E73", "#D55E00",
              "#CC79A7", "#F0E442", "#56B4E9"]
#: solarized colorsheme for axes.prop_cycle
solarized = ["#6c71c4", "#268bd2", "#2aa198", "#859900",
             "#b58900", "#cb4b16", "#dc322f", "#d33682"]

# https://blog.graphiq.com/finding-the-right-color-palettes-for-data-visualizations-fcd4e707a283
cool = gradient_to_colormap("cool",[(0, "white"), (0.25, "#DCEDC8"), (0.45, "#42B3D5"), (0.75, "#1a237e"), (1, "black")])

warm = gradient_to_colormap("warm", [(0, "white"), (0.3, "#feeb65"), (0.65, "#e4521b"), (0.85, "#4d342f"), (1, "black")])

neon = gradient_to_colormap("neon", [(0, "white"), (0.2, "#ffecb3"), (0.45, "#e85285"), (0.65, "#6a1b9a"), (1, "black")])
